# coding: utf-8
u"""Приложение для логирования запросов СМЭВ."""
import traceback
from cStringIO import StringIO

from spyne.decorator import rpc
from spyne.model import Fault
from spyne.model.primitive import Unicode
from spyne.protocol import ProtocolBase
from spyne.server.wsgi import WsgiApplication
from spyne.service import ServiceBase
from spyne_smev.server.django import DjangoApplication
from spyne_smev.fault import ApiError

from educommon.ws_log.models import SmevLog, SmevProvider
from educommon.ws_log.smev.exceptions import SpyneException


class LoggingDjangoApplication(DjangoApplication):
    u"""Переопределенный класс для логирования запросов СМЭВ."""

    def __init__(self, app, chunked=True, max_content_length=10 * 1024 * 1024,
                 block_length=8 * 1024):
        u"""Метод инициализации приложения для логирования запросов СМЭВ."""
        super(LoggingDjangoApplication, self).__init__(
            app, chunked, max_content_length, block_length)
        self.app.event_manager.add_listener('method_call', method_call)

    def generate_contexts(self, ctx, in_string_charset=None):
        u"""Call create_in_document and decompose_incoming_envelope.

        To get method_request string in order to generate contexts.
        """
        try:
            # sets ctx.in_document
            self.app.in_protocol.create_in_document(ctx, in_string_charset)

            # sets ctx.in_body_doc, ctx.in_header_doc and
            # ctx.method_request_string
            self.app.in_protocol.decompose_incoming_envelope(
                ctx, ProtocolBase.REQUEST)

            # returns a list of contexts. multiple contexts can be returned
            # when the requested method also has bound auxiliary methods.
            retval = self.app.in_protocol.generate_method_contexts(ctx)

        except Fault as e:
            log_object = ctx.transport.req['log']
            log_object.result = unicode(
                traceback.format_exc(), errors='ignore')
            ctx.in_object = None
            ctx.in_error = e
            ctx.out_error = e
            retval = (ctx,)

        return retval

    def __call__(self, request):
        u"""Логируем запрос, при вызове объекта приложения как функции."""
        request_body = request.META['wsgi.input'].read(
            request._stream.remaining)
        request.META['wsgi.input'] = StringIO(request_body)
        retval = self.HttpResponseObject()

        environ = request.META.copy()

        log_object = SmevLog()
        environ['log'] = log_object
        log_object.request = request_body
        log_object.direction = SmevLog.INCOMING
        log_object.interaction_type = SmevLog.IS_NOT_SMEV

        def start_response(status, headers):
            # Status is one of spyne.const.http
            status, reason = status.split(' ', 1)

            retval.status_code = int(status)
            for header, value in headers:
                retval[header] = value

        try:
            response = WsgiApplication.__call__(self, environ, start_response)
            self.set_response(retval, response)
        except Exception:
            log_object.result = unicode(
                traceback.format_exc(), errors="ignore")
        else:
            log_object.response = retval.content
        finally:
            if request_body:
                log_object.service_address = request.build_absolute_uri()
            if not self.is_wsdl_request(environ):
                log_object.save()

        return retval


class LoggingService(ServiceBase):
    u"""Перекрытый класс базового класса web-сервисов spyne."""

    @classmethod
    def call_wrapper(cls, ctx):
        u"""Перекрыли, чтобы отдавать валидную для СМЭВ ошибку."""
        try:
            if ctx.function is not None:
                if ctx.descriptor.no_ctx:
                    return ctx.function(*ctx.in_object)
                else:
                    return ctx.function(ctx, *ctx.in_object)

        except SpyneException as exc:
            raise ApiError(
                exc.faultcode, exc.faultstring,
                ctx.function.__name__.replace('Request', 'Response'))

        except Exception as exc:
            log = ctx.transport.req['log']
            log.result = unicode(traceback.format_exc(), errors="ignore")
            raise ApiError(
                'Server', exc.message,
                ctx.function.__name__.replace('Request', 'Response')
            )

    @rpc(Unicode, _returns=Unicode)
    def upper(self, s):
        return s.upper()


def method_call(ctx):
    u"""Обработчик события "Вызов метода".

    Словарь METHOD_VERBOSE_NAMES обязательно должен быть определен в классе
    приложения web-сервиса. Это словарь с ключами:

        * method_verbose_name - Описание сервиса.
          interaction - Вид взаимодействия (СМЭВ или не СМЭВ).
          protocol - Протокол взаимодействия, для случаев, когда вид
              взаимодействия определяется по протоколу.
          consumer_type - Тип потребителя сервиса.
          is_consumer_fio - Если True, то попытается заполнить наименование
            поставщика, из блока Man запроса.

    * - значения являются обязательными.
    """
    req_env = ctx.transport.req_env
    log = ctx.transport.req['log']

    method_info = ctx.service_class.METHOD_VERBOSE_NAMES[ctx.method_name]

    log.method_name = ctx.method_name
    log.method_verbose_name = method_info['method_verbose_name']
    log.consumer_type = method_info.get('consumer_type')

    interaction_type = method_info.get('interaction')

    if log.consumer_type == SmevLog.ENTITY:
        mnemonic = ctx.udc.in_smev_message.Service.Mnemonic
        address = req_env['HTTP_HOST'] + req_env['PATH_INFO']

        smev_providers = SmevProvider.objects.filter(
            mnemonics=mnemonic, address=address)[:1]

        if smev_providers.exists():
            log.consumer_name = smev_providers[0].mnemonics

            if interaction_type == SmevLog.IS_SMEV:
                log.source = smev_providers[0].source

            log.target_name = smev_providers[0].service_name
    else:
        if method_info.get('is_consumer_fio'):
            man = ctx.in_object.Student.Man
            fio_fields = ['Surname', 'Firstname', 'Patronymic']

            fio = u' '.join([
                getattr(man, field) for field in fio_fields
                if getattr(man, field)
            ])
            log.consumer_name = fio

    if interaction_type is None:
        protocol = method_info.get('protocol', '')
        interaction_type = (
            SmevLog.IS_SMEV if protocol.find('smev') else SmevLog.IS_NOT_SMEV)

    log.interaction_type = interaction_type
